(function( $ ){
	var easingFunctions = {
		"IN" : easeIn,
		"OUT" : easeOut,
		"INOUT" : easeInOut,
		"NONE" : easeLinear,
		"LINEAR" : easeLinear
	};
	var filters = JSManipulate;
	var animationParameters = {
		filter      : null,
		values      : null,
		stack 		: false,
		onComplete  : null,
		easing      : easeOut,
		direction   : "none",
		smoothing   : "auto",
		time        : 1000,
		steps       : 15 
	};
	var methods = {
		apply : applyFilter,
		animate : animateFilter,
		restore: restoreImage,
		convert : convertToCanvas,
		getFilters : function() {
			return filters;
		},
		getEasingFunctions : function() {
			return easingFunctions;
		},
		getAnimationParameters : function () {
			return animationParameters;
		}
	};
	function IEFallback(values){
		var prefix = "progid:DXImageTransform.Microsoft.";
		this.opacity = prefix+"Alpha(opacity="+values.amount*100+")"; 	
		this.blur = prefix+"Blur(pixelradius="+values.amount+")";
		this.grayscale = prefix+"BasicImage(grayscale=1)";
		this.invert = prefix+"BasicImage(invert=1)";
	}
 
  $.fn.jsManipulate = function(method) {
       
	var args = arguments;
	if(this.length === 0){
                
		if(method){
			if(methods[method]){
				return methods[method].apply();
			} else {
				$.error('Error: Method ' +  method + ' does not exist on jQuery.jsManipulate');
			}
		}
	} else {
               
		return this.each(function(){
			if(this.tagName === "IMG"){
                             
				if(methods[method] === convertToCanvas){
					return convertToCanvas($(this), args[1]);
				}
				var data = $(this).data("imgData");
				if(data === undefined){
					return createImageData($(this), args);
				} else if(data.originalData === undefined){
					return createImageData($(this), args);
				} else {
					return loadMethod($(this), args);
				}
			} else if(this.tagName === "CANVAS"){
                            
				return loadMethod($(this), args);
			} else {
                           
				$.error("JSIF Error: Invalid tag, only use jsManipulate on canvas or img objects.");
			}
			return this;
		});	
	}
	return this;
  };
  function supportsCanvas() {
	return !!document.createElement('canvas').getContext;
  }
  function loadMethod(canvas){
	var method = arguments[1][0];
	if(method){
		if(methods[method]){
			return methods[method].apply(canvas,Array.prototype.slice.call(arguments[1],1));
	  	} else {
	      $.error('Error: Method ' +  method + ' does not exist on jQuery.jsManipulate');
	    }
   	}
   	return null;
  }
  function convertToCanvas(el, onComplete){
  	if(el[0].tagName === "CANVAS" || !supportsCanvas()) {
  		return el;
  	}
  	var args = arguments[1];
	var imageWidth = el.attr("width");
	var imageHeight = el.attr("height");
	var canvas = $("<canvas>");
	canvas.attr({
		"width":imageWidth,
		"height":imageHeight}
	);
	var attrs = el[0].attributes;
	var attrObj = {};
	for(var i = 0; i < attrs.length; i++){
		var aName = attrs[i].nodeName;
		if(aName !== "src"){
			attrObj[aName] = attrs[i].nodeValue;
		}
	} 
	canvas.attr(attrObj);
	var context = canvas[0].getContext("2d");
	var image = new Image();
	image.src = el.attr("src");
	image.onload = function(){
		context.drawImage(image,0,0,imageWidth,imageHeight);
		canvas.data("imgData", {
			originalData: canvas[0].getContext("2d").getImageData(0,0,canvas[0].width,canvas[0].height)
		});
		canvas.data("filterValues", {});
		if(onComplete){
			onComplete(canvas);	
		}
	};
	el.replaceWith(canvas);	
	return canvas;
  }
  function createImageData(el){
  	var args = arguments[1];
  	if(!supportsCanvas()){
  		loadMethod(el,args);
  		return el;	
  	}
	var imageWidth = el.attr("width");
	var imageHeight = el.attr("height");
	var canvas = $("<canvas>");
	canvas.attr({
		"width":imageWidth,
		"height":imageHeight
		});
	var context = canvas[0].getContext("2d");
	var image = new Image();
	image.src = el.attr("src");
	image.onload = function(){
		context.drawImage(image,0,0,imageWidth,imageHeight);
		var imageData = canvas[0].getContext("2d").getImageData(0,0,canvas[0].width,canvas[0].height);
		el.data("imgData", {
			originalData: imageData
		});
		el.data("currentData", imageData);
		canvas.data("filterValues", {});
		loadMethod(el,args);
	};
	return el;
  }
  function parseEasingFunction(str){
  	if(typeof str === "string"){
		if(easingFunctions[str.toUpperCase()]){
			return easingFunctions[str.toUpperCase()];
		}
	} else if(typeof str === "function"){
		return str;
	}
	return easeOut;
  }
  //return [startX, endX, startY, endY]
  function parseDirection(str,width,height){
  	if(typeof str === "string"){
  		str = str.toUpperCase();
  		if(str !== "NONE"){
	  		if(str === "RIGHT"){
	  			return [0,width,height,height];
	  		} else if (str === "LEFT"){
	  			return [width,0,height,height];
	  		} else if (str === "DOWN"){
	  			return [width,width,0,height];
	  		} else if (str === "UP"){
	  			return [width,width,height,0];
	  		}
  		}
  	}
  	return [-1,-1,-1,-1];
  }
  function parseValue(value,current){
  	var index = 0;
  	if(typeof value === "boolean") return value;
  	return ((index = String(value).indexOf("+=")) !== -1) ? current+parseFloat(value.substring(2)) : ((index = String(value).indexOf("-=")) !== -1) ? current-parseFloat(value.substring(2)) : parseFloat(value);
  }
  function fetchOriginalData(element, xPos, yPos, width, height){
  	if(element[0].tagName === "CANVAS"){
  		if(element.data("imgData") === undefined){
  			var imageData = element[0].getContext("2d").getImageData(0,0,element[0].width,element[0].height);
	  		element.data("imgData", {
				originalData: imageData
			});
	  	}
  	}
  	var origData = element.data("imgData").originalData;
  	return getImageDataRect(origData, xPos, yPos, width, height);
  }
  function getImageDataRect(imageData, xPos, yPos, width, height){
  	if(width <= 0 || height <= 0){
  		return null;
  	}
  	var returnData = document.createElement("canvas").getContext("2d").createImageData(width,height);
  	var rPix = 0;
	for (var y = yPos; y < yPos+height; y++) {
		for (var x = xPos; x < xPos+width; x++) {
			var pixel = (y*imageData.width + x)*4;
			for(i = 0; i < 4; i++){
				returnData.data[rPix++] = imageData.data[pixel+i];
			}
		}
	}   	
    return returnData;
  }
  function getCurrentData(element){
  		if(element[0].tagName === "CANVAS"){
  			return element[0].getContext("2d").getImageData(0,0,element[0].width, element[0].height);	
  		} else {
  			return element.data("currentData");	
  		}
  }
  function insertImageData(element, data, xPos, yPos, imgContext){
	if(element[0].tagName === "CANVAS"){
		element[0].getContext("2d").putImageData(data,xPos,yPos);
	} else {
		var context = imgContext;
		var w = element[0].width;
		var h = element[0].height;
		if(context === undefined){
			var tmpCanvas = document.createElement("canvas");
			tmpCanvas.width = w;
			tmpCanvas.height = h;
			context = tmpCanvas.getContext("2d");
		}
		context.putImageData(data, xPos, yPos);
		element.data("currentData", context.getImageData(0,0,w,h));
		var uri = context.canvas.toDataURL("image/png");
		element.attr('src', uri);
	}
  }
  function parseFilter(filter){
	if(!filter){
		$.error("No filter defined");
	} else if (!filters[filter]){
		$.error("Filter: " + filter + " does not exist");
	} else {
		return filters[filter];
	}
	return null;
  }
  function restoreImage(){
  	return this.each(function(){
  		var $this = $(this);
  		var imageData = fetchOriginalData($this,0,0,$this[0].width,$this[0].height);
		if(imageData){
			insertImageData($this, imageData, 0, 0);
		}
  	});
  }
  function applyFilter(settings){
  	var filter = settings["filter"];
  	var stack = settings["stack"];
  	if(stack === undefined){
  		stack = false;	
  	}
  	var onComplete = settings["onComplete"];
  	var filterObj = parseFilter(filter);
  	var filterVals = $.extend({},filterObj.defaultValues);
	var val;
  	return this.each(function() {
  		var $this = $(this);
  		if($this.data("filterValues") === undefined){
  			$this.data("filterValues",{});
  		}
  		var isIdentical = true;
		var filterData = $this.data("filterValues");
		if(filterData[filter] === undefined){
			filterData[filter] = $.extend({},filterObj.defaultValues);
			isIdentical = false;
		}
  		for(val in settings["values"]){
  			if(settings["values"].hasOwnProperty(val)){
				filterVals[val] = parseValue(settings["values"][val],filterData[filter][val]);
				if(filterVals[val] !== filterData[filter][val]){
					isIdentical = false;
				}
			}
		}
	  	if(!supportsCanvas()){
	  		var filterString = new IEFallback(filterVals)[filter];
	  		if(filterString !== undefined) {
	  			$this.css("filter", filterString);
	  		}
	  		$this.trigger("filterComplete");
	  		if(onComplete){
	  			onComplete();	
	  		}
	  		return;
	  	}
		var numValues = 0;
		for(val in filterObj.defaultValues){
			if(filterObj.defaultValues.hasOwnProperty(val)){
				numValues++;
			}
		}
		if(isIdentical && numValues > 0 && $this.data("stacked") === false && !stack){
			$this.trigger("filterComplete");
			if(onComplete){
	  			onComplete();	
	  		}
			return;
		}
		var imageData;
		if(stack){
			imageData = getCurrentData($this);	
		} else {
			imageData = fetchOriginalData($this,0,0,$this[0].width,$this[0].height);
		}
		
		if(imageData){
			filterObj.filter(imageData,filterVals);
			insertImageData($this, imageData, 0, 0);
		}
		for(var filt in filters){
			if(filters.hasOwnProperty(filt)){
				if(filt !== filter && (filterData[filt])){
					filterData[filt] = $.extend({},filterObj.defaultValues);
				}
			}
		}
		filterData[filter] = $.extend({},filterVals);
		$this.data("stacked", stack);
		$this.trigger("filterComplete");
		if(onComplete) {
			onComplete();	
		}
  	});
  }
  
  function animateFilter(options){
  	if(!supportsCanvas()){
  		this.trigger("filterComplete");
  		if(onComplete){
  			onComplete();	
  		}
  		return this;	
  	}
	var settings = $.extend({},animationParameters);
	if(options){
		$.extend(settings,options);
	}
	var stack = settings["stack"];
	var onComplete = settings["onComplete"];
	var filter = settings["filter"];
  	var filterObj = parseFilter(filter);
  	var filterVals = $.extend({},filterObj.defaultValues);
	var easingFunc = parseEasingFunction(settings["easing"]);
	
	var totalSteps = parseInt(settings["steps"], 10);
	var totalTime = parseInt(settings["time"],10);
	var stepTime = totalTime/totalSteps;
	var val;

	return this.each(function(){
		var $this = $(this);
		if($this.data("filterValues") === undefined){
  			$this.data("filterValues",{});
  		}
  		var isIdentical = true;
		var filterData = $this.data("filterValues");
		var animationVals = {};
		if(filterData[filter] === undefined){
			filterData[filter] = $.extend({},filterObj.defaultValues);
			isIdentical = false;
		}
		for(val in settings["values"]){
			if(settings["values"].hasOwnProperty(val)){
				filterVals[val] = parseValue(settings["values"][val],filterData[filter][val]);
				if(filterVals[val] !== filterData[filter][val]){
					isIdentical = false;
				}
			}
		}
		for(val in filterVals){
			if(filterVals.hasOwnProperty(val)){
				animationVals[val] = {
					from : filterData[filter][val],
					to : filterVals[val]
				};
			}
		}
		var numValues = 0;
		for(val in filterObj.defaultValues){
			if(filterObj.defaultValues.hasOwnProperty(val)){
				numValues++;
			}
		}
		if(isIdentical && numValues > 0 && $this.data("stacked") === false && !stack){
			$this.trigger("filterComplete");
			if(onComplete){
	  			onComplete();	
	  		}
			return;
		}
		for(var filt in filters){
			if(filters.hasOwnProperty(filt)){
				if(filt !== filter && (filterData[filt])){
					filterData[filt] = $.extend({},filterObj.defaultValues);
				}
			}
		}
		var stackedData;
		if(stack){
			stackedData = getCurrentData($this);
		}
		filterData[filter] = $.extend({},filterVals);
		var smoothingPieces = (settings["smoothing"] === "auto") ? -1 : (settings["smoothing"] <= 0) ? 1 : settings["smoothing"];
		var currentStep = 0;
		var posValues = [-1,-1,-1,-1];
		if(filters[filter].isDirAnimatable){
			posValues = parseDirection(settings["direction"],$this[0].width,$this[0].height);
		}
		if(posValues[0] === -1){
			smoothingPieces = 1;
		}
		var dirP = 1;
		var dirA = 1;
		var piecePosX = 0;
		var piecePosY = 0;
		var pieceWidth = $this[0].width;
		var pieceHeight =$this[0].height;
		var prevPosX = pieceWidth;
		var prevPosY = pieceHeight;
		var prevWidth = pieceWidth;
		var prevHeight = pieceHeight;
		
		if(posValues[0] > posValues[1]){
			var tmpX = posValues[0];
			posValues[0] = posValues[1];
			posValues[1] = tmpX;
			dirP = -1;
		}
		if(posValues[2] > posValues[3]){
			var tmpY = posValues[2];
			posValues[2] = posValues[3];
			posValues[3] = tmpY;
			dirP = -1;
		}
		var horDir = (posValues[0] !== posValues[1]);
		var prevGradientWidth = null;
		var prevGradientHeight = null;
		var prevGradientPosX = null;
		var prevGradientPosY = null;
		function calcStep(){
			var step = currentStep + 1;
			 
		}
		var imageData = null;
		var imgContext = null;
		var animate;
		var animationStep = function(){
			var iterations = smoothingPieces;
			if(smoothingPieces !== 1){
				if(currentStep > 0){
					if(stack){
						imageData = getImageDataRect(stackedData, Math.floor(piecePosX),Math.floor(piecePosY),Math.ceil(pieceWidth+1),Math.ceil(pieceHeight+1));
					} else {
						imageData = fetchOriginalData($this,Math.floor(piecePosX),Math.floor(piecePosY),Math.ceil(pieceWidth+1),Math.ceil(pieceHeight+1));
					}
					if(imageData){
						filterObj.filter(imageData,filterVals);
						insertImageData($this, imageData, Math.floor(piecePosX),Math.floor(piecePosY), imgContext);
					}
				}
				if(currentStep === totalSteps){
					setTimeout(function(){
						if(stack){
							imageData = getImageDataRect(stackedData, Math.floor(piecePosX),Math.floor(piecePosY),Math.ceil(pieceWidth),Math.ceil(pieceHeight));
						} else {
							imageData = fetchOriginalData($this,Math.floor(piecePosX),Math.floor(piecePosY),Math.ceil(pieceWidth),Math.ceil(pieceHeight));
						}
						if(imageData){
							filterObj.filter(imageData,filterVals);
							insertImageData($this, imageData, Math.floor(piecePosX),Math.floor(piecePosY), imgContext);
						}
						clearInterval(animate);
					},totalTime/totalSteps);
				}
			}
			if(currentStep === totalSteps){
				clearInterval(animate);
				$this.data("stacked", stack);
				$this.trigger("filterComplete");
				if(onComplete){
					onComplete();
				}
				return;
			}
			currentStep++;
			$this.trigger("filterStep",{
				step : currentStep 
			});
			var currentTime = (totalTime/totalSteps)*currentStep;
			var value, from, to;
			if(posValues[0] === -1){
				for(value in animationVals){
					if(animationVals.hasOwnProperty(value)){
						from = animationVals[value].from;
						to = animationVals[value].to;
						if(from > to){
							filterVals[value] = from - easingFunc.apply(this,[currentTime,0,from-to,totalTime]);
						} else {
							filterVals[value] = from + easingFunc.apply(this,[currentTime,0,to-from,totalTime]);
						}
					}
				}
			} else {
				if(horDir){
					if(dirP === 1){
						piecePosX = Math.floor(easingFunc.apply(this,[currentTime-stepTime,posValues[0],posValues[1],totalTime]));
						pieceWidth = Math.floor(easingFunc.apply(this,[currentTime,posValues[0],posValues[1],totalTime])) - piecePosX;
					} else if(dirP === -1){
						piecePosX = posValues[1] - Math.ceil(easingFunc.apply(this,[currentTime,posValues[0],posValues[1],totalTime]));
						pieceWidth = prevPosX - piecePosX;
					}	
				} else {
					if(dirP === 1){
						piecePosY = Math.ceil(easingFunc.apply(this,[currentTime-stepTime,posValues[2],posValues[3],totalTime]));
						pieceHeight = Math.ceil(easingFunc.apply(this,[currentTime,posValues[2],posValues[3],totalTime])) - piecePosY;
					} else if(dirP === -1){
						piecePosY = posValues[3] - Math.ceil(easingFunc.apply(this,[currentTime,posValues[2],posValues[3],totalTime]));
						pieceHeight = prevPosY - piecePosY;
					}
				}
			}
			if(pieceWidth > 0 && pieceHeight > 0){
				iterations = smoothingPieces;
				if(iterations === -1){
					iterations = (horDir) ? Math.ceil(pieceWidth/2) + 5 : Math.ceil(pieceHeight/2) + 5;
				}
				var gradientWidth = (horDir) ? pieceWidth/iterations : pieceWidth;
				var gradientHeight = (!horDir) ? pieceHeight/iterations : pieceHeight;
				
				for(var i = 0; i < iterations; i++){
					if(Math.ceil(gradientWidth) > 0 && Math.ceil(gradientHeight) > 0){
						var gradientVals = {}; 
						if(iterations === 1){
							gradientVals = filterVals;
						} else {
							for(value in animationVals){
								if(animationVals.hasOwnProperty(value)){
									from = animationVals[value].from;
									to = animationVals[value].to;
									var dirV = 1;
									if(from > to){
										dirV = -1;
									} 
									gradientVals[value] = ((to-from))/iterations;
									if(dirP === 1){
										gradientVals[value] *= i; 
									} else {
										gradientVals[value] *= iterations-i;
									}
									gradientVals[value] = to - gradientVals[value];
								}
							}
						}
						var optimizable = (filterVals.radius !== undefined && filterVals.centerY !== undefined && filterVals.centerX !== undefined);
						var optimizedSize = false;
						if(optimizable){
							if(!prevGradientWidth){
								 prevGradientWidth = animationVals["radius"].from * 2;
							}
							if(!prevGradientHeight){
								 prevGradientHeight = animationVals["radius"].from * 2;
							}
							gradientWidth = gradientVals.radius * 2;
							gradientHeight = gradientVals.radius * 2;
							if(prevGradientWidth > gradientWidth) {
								gradientWidth = prevGradientWidth;
							}
							if(prevGradientHeight > gradientHeight) {
								gradientHeight = prevGradientHeight;
							}
							prevGradientWidth = gradientVals.radius * 2;
							prevGradientHeight = gradientVals.radius * 2;
							optimizedSize = true;
						}
 						var gradientPosX = (horDir) ? piecePosX+gradientWidth*i : piecePosX;
						var gradientPosY = (!horDir) ? piecePosY+gradientHeight*i : piecePosY; 
						if(optimizable){
							gradientPosX = pieceWidth*gradientVals.centerX - gradientWidth/2;
							gradientPosY = pieceHeight*gradientVals.centerY - gradientHeight/2;
							var tmpA = gradientPosX;
							var tmpB = gradientPosY;
							if(!prevGradientPosX) {
								 prevGradientPosX =  pieceWidth*animationVals["centerX"].from - gradientWidth/2;
							}
							if(!prevGradientPosY) {
								 prevGradientPosY = pieceHeight*animationVals["centerY"].from - gradientWidth/2;
							}
							gradientWidth += Math.abs(gradientPosX-prevGradientPosX)*2;
							gradientHeight += Math.abs(gradientPosY-prevGradientPosY)*2;
							gradientPosX -= Math.abs(gradientPosX-prevGradientPosX);
							gradientPosY -= Math.abs(gradientPosY-prevGradientPosY);
							if(prevGradientPosX < gradientPosX){
								gradientPosX = prevGradientPosX;
							}
							if(prevGradientPosY < gradientPosY){
								gradientPosY = prevGradientPosY;
							}
							prevGradientPosX = tmpA;
							prevGradientPosY = tmpB;
							gradientVals.centerX = 0.5;
							gradientVals.centerY = 0.5;
						}
						if(optimizedSize && currentStep === 1){
							if(stack){
								imageData = getImageDataRect(stackedData, 0,0,$this[0].width, $this[0].height);
							} else {
								imageData = fetchOriginalData($this,0,0,$this[0].width, $this[0].height);
							}
							
							insertImageData($this,imageData,0,0,$this[0].width,$this[0].height);	
						}
						if(gradientPosX < 0){
							gradientVals.centerX += gradientPosX / gradientWidth;
							gradientPosX = 0;
						} else if(gradientPosX + gradientWidth >= $this[0].width){
							gradientVals.centerX -= ($this[0].width - (gradientPosX + gradientWidth)) / gradientWidth;
							gradientWidth += $this[0].width - (gradientPosX + gradientWidth) - 1;
						}
						if(gradientPosY < 0){
							gradientVals.centerY += gradientPosY / gradientHeight;
							gradientPosY = 0;
						} else if(gradientPosY + gradientHeight >= $this[0].height){
							gradientVals.centerY -= ($this[0].height - (gradientPosY + gradientHeight)) / gradientWidth;
							gradientHeight += $this[0].height - (gradientPosY + gradientHeight) - 1;
						}
						
						if(posValues[0] === -1 && filterVals.radius === undefined){
							if(stack){
								imageData = getImageDataRect(stackedData,0,0,pieceWidth,pieceHeight);
							} else {
								imageData = fetchOriginalData($this,0,0,pieceWidth,pieceHeight);
							}
						} else {
							if(stack){
								imageData = getImageDataRect(stackedData, Math.floor(gradientPosX),Math.floor(gradientPosY),Math.ceil(gradientWidth)+1,Math.ceil(gradientHeight)+1);
							} else {
								imageData = fetchOriginalData($this,Math.floor(gradientPosX),Math.floor(gradientPosY),Math.ceil(gradientWidth)+1,Math.ceil(gradientHeight)+1);
							}
						}
						if(imageData){
							filterObj.filter(imageData,gradientVals);
							insertImageData($this, imageData, Math.floor(gradientPosX),Math.floor(gradientPosY), imgContext);
						}
					}
				}
			}
			prevPosX = piecePosX;
			prevPosY = piecePosY;
			prevWidth = pieceWidth;
			prevHeight = pieceHeight;
		};
		if(this.tagName === "CANVAS"){
			animate = setInterval(animationStep,totalTime/totalSteps);
		} else {
			var tmpCanvas = document.createElement("canvas");
			tmpCanvas.width = $this.width;
			tmpCanvas.height = $this.height;
			imgContext = tmpCanvas.getContext("2d");
			var img = new Image();
			img.src = $this.attr('src');
			img.onload = function (){
				imgContext.drawImage(img,0,0,$this[0].width, $this[0].height);
				animate = setInterval(animationStep,totalTime/totalSteps);
			};
		}
	});
  }
  
  //EASING FUNCTIONS
  function easeLinear(t,b,c,d){
  	return b+(t/d)*c;
  }
  function easeIn (t, b, c, d) {
	return c*(t/=d)*t + b;
  }
  function easeOut(t, b, c, d) {
	return -c *(t/=d)*(t-2) + b;
  }
  function easeInOut(t, b, c, d) {
	if ((t/=d/2) < 1){
		 return c/2*t*t + b;
	}
	t--;
	return -c/2 * ((t)*(t-2) - 1) + b;
  }
})( jQuery );			